 /*
 * Created on 6 sept. 2003
 *
 */
package org.gudy.azureus2.irc;

//TODO Add stuff to MessageBuddles.
//TODO Comment the INFOs... BAH!

import org.jibble.pircbot.Colors;
import org.jibble.pircbot.PircBot;
import org.jibble.pircbot.User;
import org.gudy.azureus2.core3.util.Constants;
import org.gudy.azureus2.plugins.*;
import org.gudy.azureus2.plugins.download.Download;

import java.net.*;
import java.io.*;
import java.util.Arrays;
import java.util.Locale;

import com.aelitis.azureus.core.AzureusCoreFactory;
import com.aelitis.azureus.core.dht.DHT;
import com.aelitis.azureus.plugins.dht.DHTPlugin;

import org.gudy.azureus2.plugins.network.ConnectionManager;
import org.gudy.azureus2.plugins.utils.*;


/**
 * @author Olivier
 * 
 */

public class 
IrcClient 
	extends PircBot 
{
	public static final String	CONFIG_IRC_SERVER 			= "Irc Server";
	public static final String	CONFIG_IRC_SERVER_DEFAULT	= "irc.freenode.net";
	
	public static final String	CONFIG_IRC_CHANNEL			= "Irc Channel";
	public static final String	CONFIG_IRC_CHANNEL_DEFAULT	= "";
	
	public static final String	CONFIG_IRC_USER				= "Irc Login";
	public static final String	CONFIG_IRC_USER_DEFAULT		= "";
	
	//The settings for the INFO retrieval.
	public static final String	CONFIG_IRC_SEND_USER_INFO				= "Irc User Info";
	public static final boolean	CONFIG_IRC_SEND_USER_INFO_DEFAULT		= true;
	
	//And the HTML log feature.
	public static final String	CONFIG_IRC_LOG				= "Irc Log";
	public static final boolean	CONFIG_IRC_LOG_DEFAULT		= false;
	
	  //Some constants to allow helpers to only retrieve the INFO they need.
	  final static public int REQUESTED_JAVA = 1;
	  final static public int REQUESTED_SETTINGS = 1<<1;
	  final static public int REQUESTED_ACTIVE = 1<<2;
	  final static public int REQUESTED_STATUS = 1<<3;
	  final static public int REQUESTED_NETWORK = 1<<4;
	  final static public int REQUESTED_MISC = 1<<5;
	  final static public int REQUESTED_PLUGINS = 1<<6;
	
	PluginInterface		plugin_interface;
	PluginConfig		plugin_config;
	PluginManager		plugin_manager;
	LocaleUtilities 	locale_utils;
	
	public String chanTopic; //Used to store the topic so we can use it.
	boolean connected = false; //This is just a once use thing for the nick correction.
	String srvName;
	String channel;
	String azVersion;
	IrcListener listener;
	String userName;
	String static_id;
	//This is the default error pre-fix.
	String catchError = "Something went horribly wrong... Please notify another helper (Error: ";
	String lCatchError = "Something went horribly wrong... Please notify a helper (Error: ";
	ConnectionManager	connection_manager;
	
	public
	IrcClient(
			PluginInterface	_plugin_interface,
			IrcListener 	_listener)
	{
		//Shame-less self advertising.
		super.setVersion("Azureus IRC Plugin 2.00");
		super.setFinger("Azureus IRC Plugin 2.00");
		plugin_interface	= _plugin_interface;
		listener = _listener;
		locale_utils = plugin_interface.getUtilities().getLocaleUtilities();
		plugin_config = plugin_interface.getPluginconfig();
		plugin_manager = plugin_interface.getPluginManager();
		azVersion = plugin_interface.getAzureusVersion();
		
		srvName 	= plugin_config.getPluginStringParameter(CONFIG_IRC_SERVER, CONFIG_IRC_SERVER_DEFAULT );
		channel 	= plugin_config.getPluginStringParameter(CONFIG_IRC_CHANNEL, CONFIG_IRC_CHANNEL_DEFAULT );
		userName 	= plugin_config.getPluginStringParameter(CONFIG_IRC_USER, CONFIG_IRC_USER_DEFAULT);
		static_id = plugin_config.getStringParameter( "ID", "        " ).substring( 0, 8 );  //first chars of the client instance id
		
		//Auto-correct the user.
		channel = channel.trim();

		//Get the default channel if they don't speficy one..
		if(channel.equals("")) {
			channel = locale_utils.getLocalisedMessageText("IrcClient.defaultChannel");
		}
		
		if(userName.equals("")) {
			listener.systemMessage(locale_utils.getLocalisedMessageText("IrcClient.noNick"));
			return;
		}
		else if (userName.equals(userName.toUpperCase())) {
			userName = userName.toLowerCase(); //Caps are not fun.
		}
		
		String az_version = azVersion;
		az_version = az_version.replaceAll("_", "" );
		az_version = az_version.replaceAll("\\.", "" );
		
		PluginManager	pm = AzureusCoreFactory.getSingleton().getPluginManager();
		connection_manager = pm.getDefaultPluginInterface().getConnectionManager();

		setName( userName ); //also sets NICK
		setLogin( static_id );
		setVersion( "AZ" +az_version+ " | " +static_id+ " | " +getVersion() );
		
		Thread t = new Thread() {
			public void run() {			  
				try {
					listener.systemMessage("");
					listener.systemMessage(locale_utils.getLocalisedMessageText("IrcClient.copyright"));
					listener.systemMessage("");
					//Thread starter was here.
					listener.systemMessage(locale_utils.getLocalisedMessageText("IrcClient.connecting") + " " + srvName);
					connect(srvName);
					listener.systemMessage(locale_utils.getLocalisedMessageText("IrcClient.connected") + " " + srvName);
					connected = true; //Enable the nick correction from now.					
					listener.systemMessage(locale_utils.getLocalisedMessageText("IrcClient.joining") + " " + channel);
					joinChannel(channel);
					listener.systemMessage(locale_utils.getLocalisedMessageText("IrcClient.channel") + " " + channel + " " +  locale_utils.getLocalisedMessageText("IrcClient.joined"));
				}
				catch (Exception e) {
					listener.systemMessage(locale_utils.getLocalisedMessageText("IrcClient.error") + ": " + e.getMessage());
				}
			}
		};
		t.setDaemon(true);
		t.start();
	}
	
	//URL retriever was here.
	
	protected void onMessage(String channel, String sender, String login, String hostname, String message) {
		if(listener != null) {
			listener.messageReceived(sender,Colors.removeFormattingAndColors(message));
		}
	}
	
	public void close() {
		super.quitServer( plugin_interface.getAzureusName() + " " + plugin_interface.getAzureusVersion());
		try {
			super.dispose();
		}
		catch (Exception e) { 
			
		}
	}
	
	public void sendMessage(String message) {
			listener.messageReceived(userName,message);
			super.sendMessage(channel,message);	
	}
	
	//This replaces the normal .sendMessage (Which goes via a queue).
	//The user is less likely to flood themselves and it fixes an INFO problem
	//Where it would fill the users queue, muting them for x minutes...
	public void sendRawMessage(String target, String message)
	{
		//Basic checks for idiocy.
		if ((message.startsWith("!")) || (message.startsWith("@"))) {
			listener.notice("WARNING", "Do not use ! and @ triggers here, there are NO files - read the topic and http://azureus.aelitis.com/wiki/index.php/Rules_for_IRC");
		}
		else
		{
			//Spam users who fail to read.
			if (target.equalsIgnoreCase("#Azureus-Users"))
				listener.notice("WARNING", "THIS IS A MODERATED CHANNEL - YOU CANNOT CHAT HERE - PLEASE SELECT A HELP CHANNEL (DOUBLE CLICK) FROM THE LIST:\r\nAvailable Channels: English: type /join #Azureus-Support  Polski: wpisz /join #Azureus-Polish  Espanol: escribe /join #Azureus-Espanol  Deutsch: schreibe /join #Azureus-German  Francais: tapez /join #Azureus-French  Finnish: /join #Azureus-Finnish  Portugues: escreva /join #Azureus-Portugues  Norwegian: skriv /join #Azureus-Norwegian  Mac OSX: type /join #azureus-macosx");
			else {
			//Stop the user shouting so much.
			if (message.equals(message.toUpperCase()) && message.length() > 5)
				message = message.toLowerCase();
			}
			//Fix related to the sendMessage change.
			if (target.startsWith("#"))
				listener.messageReceived(userName, message);
			super.sendRawLine("PRIVMSG " + target + " :" + message);
		}
	}
	
	//MD5 and Filering was here.
	
	protected void onJoin(String channel, String sender, String login, String hostname) {
		listener.systemMessage(sender + " " + locale_utils.getLocalisedMessageText("IrcClient.hasjoined")  + (sender.equals(userName)?" " + channel:""));
		listener.clientEntered(sender);
	}
  
	protected void onKick(String channel, String kickerNick, String kickerLogin, String kickerHostname, String recipientNick, String reason) {
		listener.systemMessage(kickerNick + " " + locale_utils.getLocalisedMessageText("IrcClient.haskicked") + " " + recipientNick + " (" + reason + ").");
		listener.clientExited(recipientNick);
		//Tell them they got booted, why, a wiki link, and set the channel.
		if (recipientNick.equals(userName)) {
			listener.notice("WARNING", "You are no longer in the channel as you have been kicked (" + reason + ") Please respect the rules... http://azureus.aelitis.com/wiki/index.php/Rules_for_IRC");
			this.channel = "Kicked from " + channel;
			listener.allExited();
		}
	}
	
	protected void onQuit(String sourceNick, String sourceLogin, String sourceHostname, String reason) {
		listener.systemMessage(sourceNick + " " + locale_utils.getLocalisedMessageText("IrcClient.hasleft") + " (" + reason + ").");
		listener.clientExited(sourceNick);
	}
  
	protected void onPart(String channel, String sender, String login, String hostname) {
		listener.systemMessage(sender + " " + locale_utils.getLocalisedMessageText("IrcClient.hasleft")  + (sender.equals(userName)?" " + channel:""));
		listener.clientExited(sender);
	}
  
	protected void onNickChange(String oldNick, String login, String hostname, String newNick) {
		listener.systemMessage(oldNick + " " + locale_utils.getLocalisedMessageText("IrcClient.nowknown") + " " + newNick);
		listener.clientExited(oldNick);
		listener.clientEntered(newNick);
	}
	
	protected void onAction(String sender, String login, String hostname, String target, String action) {
		listener.action(sender,action);
	}
  
	public String getUserName() {
		return userName;
	}

	public void setUserName(String _userName) {
		_userName = _userName.trim();
		//This will prevent the user being l33t. Well, at least the first char...		
		if (! _userName.startsWith("`") && ! _userName.startsWith("") && ! _userName.startsWith("|") && ! _userName.startsWith("[") && ! _userName.startsWith("]") && ! _userName.startsWith("_") && ! _userName.startsWith("^")){
			if (_userName.equals(_userName.toUpperCase())) {
				_userName = _userName.toLowerCase();
			}
		}
		else {
			//Tell them the 'reason'.
			listener.notice("WARNING", "Your nick was not changed as it would have made responding to you more difficult, Please choose one containing only alphanumerc characters... http://azureus.aelitis.com/wiki/index.php/Rules_for_IRC");
		}
		userName = _userName;
		changeNick(userName);
	}

	public void sendAction(String action) {
		super.sendAction(channel,action);
	}
  
	protected void onUserList(String channel, User[] users) {
		if(! this.channel.equals(channel)) {
			return;
		}
	  for(int i = 0 ; i < users.length ;i++) {
		  listener.clientEntered(users[i].getNick());
		}
	}
	
  protected void onPrivateMessage(String sender, String login, String hostname, String message) {
	  //No messages unless the user is an Op or voice... 
	  //Commented out because helpers might use this to get a user BACK from a channel they don't have Ops in.
	  //if (isChannelVoice(sender) || isChannelOp(sender)) {
		  listener.privateMessage(sender, message, hostname);
		  }
	  //}
  protected void onNotice(String sourceNick, String sourceLogin, String sourceHostname, String target, String notice) {
	  //No messages unless the user is an Op or voice... 
	  //Commented out because helpers might use this to get a user BACK from a channel they don't have Ops in.
	  //if (isChannelVoice(sourceNick) || isChannelOp(sourceNick))
	  //{
		  //}
		  //Special colours for the Services...
		  if (sourceNick.equals("NickServ") || (sourceNick.equals("ChanServ")))
		  {
			  listener.systemMessage(notice);
		  } else {
			  listener.notice(sourceNick,notice);
		  }
	  } 
  
  protected void onServerResponse(int code, String response) {
	  switch (code)
	  {
	  case ERR_NOSUCHNICK:
		  listener.systemMessage(response);
		  break;
		  //Fancy auto nick corrector.
	  case ERR_NICKNAMEINUSE:
		  if (connected) {
		  listener.systemMessage(response);
		  response = response.substring(response.indexOf(" ") + 1);
		  response = response.substring(0, response.indexOf(" "));
		  this.setUserName(response + "_");
		  }
		  break;
		  //User got banned, spam them a wiki link.
	  case ERR_BANNEDFROMCHAN:
		  listener.notice("WARNING", response + " See: http://azureus.aelitis.com/wiki/index.php/Rules_for_IRC");
		  break;
		  //Hell, this might save a user or two from wearing out their keyboard...
	  case ERR_CANNOTSENDTOCHAN:
		  listener.notice("WARNING", response);
		  break;
		  }
	  }
  
  protected void onTopic(String channel, String topic, String setBy, long date, boolean changed) {
	  chanTopic = topic; //Set the topic, used latter.
	  listener.systemMessage(locale_utils.getLocalisedMessageText("IrcClient.topicforchannel") + " " + channel + " : " + topic);
	  }
  
  protected void onDisconnect() {
	  listener.systemMessage(locale_utils.getLocalisedMessageText("IrcClient.disconnected") + " " + srvName);
	  }
  
  public String getChannel() {
	  return channel;
	  }

  public String getSrvName() {
	  return srvName;
	  }
  
  public void changeChannel(String channel, String message) {
	  //Is the user trying to join the same channel? Tell them off and ignore it.
	  //Or, are they not... Well then do some checks and complete the change.
	  if (! channel.equalsIgnoreCase(this.channel)) {
		  channel = channel.trim();
		  //The message is the channel they are going to... And how...
		  partChannel(this.channel + " :" + message);
		  User[] users = super.getUsers(this.channel);
		  for(int i=0 ; i< users.length ; i++) {
			  listener.clientExited(users[i].getNick());
			  }
		  this.channel = channel;
		  joinChannel(this.channel);
		  }
	  else {
		  //Tell the user off.
		  listener.notice("Information", "You're already in that channel (" + channel +")");
		  }
	  }
  
  //Used for INFO and other control functions.
  public boolean isChannelOp(String nick)
  {
    User[] users = getUsers(channel);
    int i;
    for (i = 0; i < users.length; i++)
	if (users[i].getNick().equalsIgnoreCase(nick) && users[i].isOp())
	    return true;
    return false;
  }
  //Used for INFO and other control functions.
  public boolean isChannelVoice(String nick)
  {
    User[] users = getUsers(channel);
    int i;
    for (i = 0; i < users.length; i++)
	if (users[i].getNick().equalsIgnoreCase(nick) && users[i].hasVoice())
	    return true;
    return false;
  }
  
  //INFO stuff, two functions, one for remote triggers (When a helper sends a request) and
  //			one for local (When the user types /info). There are about 25 try { } catch { }
  //			combos for each, which should prevent any fatal errors >_>
  //			And if you think I'm commenting it... HAHAHA.
  
  public void localInfo() {
	  try {
		  String UPnPEnabled;
		  String UserSelectedMode;
		  String BindToIP = "N/A";
				  try {				  
					  String patchLevel = " " + System.getProperty("sun.os.patch.level") + " ";
					  if (patchLevel == "  ") {
						  patchLevel = " ";
					  }
					  listener.systemMessage( "Info for " + getUserName() + " (" + static_id +") in " + channel +" - " + azVersion + " installed in " +  System.getProperty("user.dir") + " - " + System.getProperty("os.name") + patchLevel + System.getProperty("os.version") + " " + System.getProperty("os.arch"));
					  }
					  catch (Exception ex) {
						  listener.systemMessage( lCatchError + ex.getMessage().toString() + ")");
					  }
				  try {
					  String filedir = "";
					  try {
						  try {	
							  try {
								  try {
						  File dir = new File(System.getProperty("java.home") + File.separator + ".." + File.separator + (Constants.isOSX?".."+File.separator:""));
						    String[] children = dir.list();
					  int found = 0;
					  if (children != null)
					  {
						  for (int i=0; i<children.length; i++) {
							  if (children[i].startsWith("j")) {
							  filedir = children[i] + ", " + filedir;
							  found++;
							  }
							  }
						  }
					  filedir = filedir.substring(0,filedir.length() - 2);
					  filedir = " (" + found + "): " + filedir + " |";
					  }
						  catch (Exception ex) {
							  File dir = new File(System.getProperty("java.home") + File.separator + ".." + File.separator + (Constants.isOSX?".."+File.separator:"") + (Constants.isLinux?".."+File.separator:""));
							    String[] children = dir.list();
						  int found = 0;
						  if (children != null)
						  {
							  for (int i=0; i<children.length; i++) {
								  if (children[i].startsWith("j")) {
								  filedir = children[i] + ", " + filedir;
								  found++;
								  }
								  }
							  }
						  filedir = filedir.substring(0,filedir.length() - 2);
						  filedir = " (" + found + "): " + filedir + " |";
						  }
						  }
						  catch (Exception ex) {
							  File dir = new File(System.getProperty("java.home") + File.separator + ".." + File.separator + (Constants.isOSX?".."+File.separator:""));
							    String[] children = dir.list();
						  int found = 0;
						  if (children != null)
						  {
							  for (int i=0; i<children.length; i++) {
								  filedir = children[i] + ", " + filedir;
								  found++;
								  }
							  }
						  filedir = filedir.substring(0,filedir.length() - 2);
						  filedir = " (" + found + "): " + filedir + " |";
						  }
					  } catch (Exception ex) {
						  File dir = new File(System.getProperty("java.home") + File.separator + ".." + File.separator + (Constants.isOSX?".."+File.separator:"")+ (Constants.isLinux?".."+File.separator:""));
						    String[] children = dir.list();
					  int found = 0;
					  if (children != null)
					  {
						  for (int i=0; i<children.length; i++) {
							  filedir = children[i] + ", " + filedir;
							  found++;
							  }
						  }
					  filedir = filedir.substring(0,filedir.length() - 2);
					  filedir = " (" + found + "): " + filedir + " |";
						  }
				  } catch (Exception ex) {
					  filedir = lCatchError + ex.getMessage().toString();
					  }
				  String JLine = "";
				  if (Constants.isWindows) {
					  try {
						  Process p = Runtime.getRuntime().exec("REG QUERY \"HKLM\\SOFTWARE\\JavaSoft\\Java Update\\Policy\\\" /s");
						  InputStream is = p.getInputStream();
						  InputStreamReader isr = new InputStreamReader(is);
						  BufferedReader br = new BufferedReader(isr);
						  String line;
						  while ((line = br.readLine()) != null) {
							  if (line.indexOf("EnableJavaUpdate") > -1) {
								  if (line.indexOf("0x0") > -1) {
									  JLine = "Updater: Disabled | ";
									  }
								  else  {
									  JLine = "Updater: Enabled | ";
									  }
								  } 
							  else if (line.indexOf("EnableAutoUpdateCheck") > -1) {
								  if (line.indexOf("0x0") > -1) {
									  JLine = JLine + "Auto update check: Disabled";
									  }
								  else	{
									  JLine = JLine + "Auto update check: Enabled";
									  }
								  }
							  }
						  br.close();
						  }
					  catch (Exception Ex) {
						  
					  }
					  }
				  listener.systemMessage( "Java: " + System.getProperty("java.version") + " (" + System.getProperty("java.vendor") + ") installed in " + System.getProperty("java.home") + " | Installations" + filedir + " " + JLine);
				  }
				  catch (Exception ex) {
					  listener.systemMessage( lCatchError + ex.getMessage().toString() + ")");
				  }
				  try {
				  if (! plugin_config.getStringParameter("Bind IP","").equals("")) {
					  BindToIP = plugin_config.getStringParameter("Bind IP","");
					  }
				  String portUDP = "";
				  if (plugin_config.getIntParameter("Plugin.DHT.dht.portdefault") == 1) {
					  portUDP = plugin_config.getIntParameter("TCP.Listen.Port",6881) + "";
					  } 
				  else {
					  portUDP = plugin_config.getIntParameter("Plugin.DHT.dht.port") + "";
					  }
				  String pTLine = "Disabled";
				  String pPLine = "Disabled";
				  if (plugin_config.getIntParameter("Enable.Proxy",0) == 1)
				  {
					  pTLine = "Enabled";
				  }
				  if (plugin_config.getIntParameter("Proxy.Data.Enable",0) == 1)
				  {
					  pPLine = "Enabled";
				  }
				  String enLine = "Disabled";
				  if (plugin_config.getIntParameter("network.transport.encrypted.require",0) == 1)
				  {
					  enLine = "Enabled";
					  if (plugin_config.getStringParameter("network.transport.encrypted.min_level","RC4") == "RC4")
					  {
						  enLine += " | Minimum: RC4";
					  }
					  else 
					  {
						  enLine += " | Minimum: Plain";
					  }
					  if (plugin_config.getIntParameter("network.transport.encrypted.fallback.incoming",0) == 1)
					  {
						  enLine += " | Incoming fallback: Enabled";
					  }
					  else 
					  {
						  enLine += " | Incoming fallback: Disabled";
					  }
					  if (plugin_config.getIntParameter("network.transport.encrypted.fallback.outgoing",0) == 1)
					  {
						  enLine += " | Outgoing fallback: Enabled";
					  }
					  else 
					  {
						  enLine += " | Outgoing fallback: Disabled";
					  }
				  }
				  listener.systemMessage( "Connection: TCP Port: " + plugin_config.getIntParameter("TCP.Listen.Port",6881) + " | UDP Port: " + portUDP + " | Max connection attempts: " + plugin_config.getIntParameter("network.max.simultaneous.connect.attempts",8) + " | Bind to local IP: " + BindToIP + " | Encryption: " + enLine + " | Proxy tracker: " + pTLine + " | Proxy Data: " + pPLine);
				  }
				  catch (Exception ex) {
					  listener.systemMessage( lCatchError + ex.getMessage().toString() + ")");
				  }
				  try {
				  String upSeedingOnly = "";
				  if (plugin_config.getIntParameter("enable.seedingonly.upload.rate",0) == 0) {
					  upSeedingOnly = "N/A";
					  }
				  else  {
					  upSeedingOnly = "" + plugin_config.getIntParameter("Max Upload Speed Seeding KBs", 0);
					  }
				  listener.systemMessage( "Transfer: Max upload speed: " + plugin_config.getIntParameter("Max Upload Speed KBs", 0) + " | Max upload speed while seeding: " + upSeedingOnly + " | Max download speed: " + plugin_config.getIntParameter("Max Download Speed KBs", 0) + " | Default max upload slots: " + plugin_config.getIntParameter("Max Uploads", 4));
				  }
				  catch (Exception ex) {
					  listener.systemMessage( lCatchError + ex.getMessage().toString() + ")");
				  }
				  try {
				  listener.systemMessage( "Transfer: Max connections per torrent: " + plugin_config.getIntParameter("Max.Peer.Connections.Per.Torrent", 80) + " | Max connections globally: " + plugin_config.getIntParameter("Max.Peer.Connections.Total", 400));
				  }
				  catch (Exception ex) {
					  listener.systemMessage(lCatchError + ex.getMessage().toString() + ")");
				  }
				  try {
					  String sameIPPeers = "Disabled";
					  String prioPiece = "Disabled";
					  String lazy = "Disabled";
					  String ignorePorts = "" + plugin_config.getStringParameter("Ignore.peer.ports","None");
					  if (plugin_config.getIntParameter("Allow Same IP Peers", 0) == 1)
					  {
						  sameIPPeers = "Enabled";
					  }
					  if (plugin_config.getIntParameter("Prioritize First Piece", 0) == 1)
					  {
						  prioPiece = "Enabled";
					  }
					  if (plugin_config.getIntParameter("Use Lazy Bitfield", 0) == 1)
					  {
						  lazy = "Enabled";
					  }
					  if (ignorePorts.equals(""))
					  {
						  ignorePorts = "None";
					  }
					  listener.systemMessage("Transfer: Same IP peers: " + sameIPPeers + " | Prioritise first piece: " + prioPiece + " | Lazy bitfield: " + lazy + " | Ignore ports: " + ignorePorts);
					  }
					  catch (Exception ex) {
						  listener.systemMessage(lCatchError + ex.getMessage().toString() + ")");
					  }
				  try {
				  String numberSeedingOnly = "";
				  if (plugin_config.getIntParameter("StartStopManager_bMaxActiveTorrentsWhenSeedingEnabled", 0) == 0) {
					  numberSeedingOnly = "N/A";
					  }
				  else  {
					  numberSeedingOnly = "" + plugin_config.getIntParameter("StartStopManager_iMaxActiveTorrentsWhenSeeding");
					  }
				  String supers = "";
				  if (plugin_config.getIntParameter("Use Super Seeding", 0) == 0)
					  supers = "Disabled";
				  else
					  supers = "Enabled";
				  listener.systemMessage("Queue: Max downloads: " + plugin_config.getIntParameter("max downloads", 4) + " | Max torrents: " + plugin_config.getIntParameter("max active torrents", 4) + " | Max torrents while seeding: " + numberSeedingOnly + " | Super seeding: " + supers);
				  }
				  catch (Exception ex) {
					  listener.systemMessage(lCatchError + ex.getMessage().toString() + ")");
				  }
				  try {
					  Download[] managers = plugin_interface.getDownloadManager().getDownloads();
					  int forced = 0;
					  int downloading = 0;
					  int downloading_f = 0;
					  int seeding = 0;
					  int seeding_f = 0;
					  int queued = 0;
					  int queued_f = 0;
					  int erroring = 0;
					  int erroring_f = 0;
					  int preparing = 0;
					  int preparing_f = 0;
					  int stopped = 0;
					  int stopped_f = 0;
					  int ready = 0;
					  int ready_f = 0;
					  int stoping = 0;
					  int stoping_f = 0;
					  int waiting = 0;
					  int waiting_f = 0;
				      for (int i = 0; i < managers.length; i++) {
				    	  Download manager = managers[i];
				    	  if (manager.isForceStart())
				    		  forced++;
				          int state = manager.getState();
				          if (state == Download.ST_DOWNLOADING)
				          {
				            downloading++;
				            if (manager.isForceStart())
				            	downloading_f++;
				          }
				          else if (state == Download.ST_SEEDING)
				          {
				            seeding++;
				            if (manager.isForceStart())
				            	seeding_f++;
				          }
				          else if (state == Download.ST_QUEUED)
				          {
				            queued++;
				            if (manager.isForceStart())
				            	queued_f++;
				          }
				          else if (state == Download.ST_ERROR)
				          {
				        	  erroring++;
				        	  if (manager.isForceStart())
					            	erroring_f++;
					          }
				          else if (state == Download.ST_PREPARING)
				          {
				        	  preparing++;
			        	  if (manager.isForceStart())
			        		  preparing_f++;
				          }
				          else if (state == Download.ST_STOPPED)
				          {
				        	  stopped++;
			        	  if (manager.isForceStart())
				            	stopped_f++;
				          }
				          else if (state == Download.ST_READY)
				          {
				        	  ready++;
				        	  if (manager.isForceStart())
					            	ready_f++;
					          }
				          else if (state == Download.ST_STOPPING)
				          {
				        	  stoping++;
				        	  if (manager.isForceStart())
					            	stoping_f++;
					          }
				          else if (state == Download.ST_WAITING)
				          {
				        	  waiting++;
				        	  if (manager.isForceStart())
					            	waiting_f++;
					          }
				        }
				      listener.systemMessage("Active: Total: " + managers.length + " (F: " + forced + ") | Downloading: " + downloading + " (F: " + downloading_f + ") | Seeding: " + seeding + " (F: " + seeding_f + ") | Queued: " + queued + " (F: " + queued_f + ") | Stopped: " + stopped + " (F: " + stopped_f + ") | Ready: " + ready + " (F: " + ready_f + ") | Waiting: " + waiting + " (F: " + waiting_f + ") | Checking: " + preparing + " (F: " + preparing_f + ") | Stopping: " + stoping + " (F: " + stoping_f + ") | Errored: " + erroring + " (F: " + erroring_f + ")");
				  }
				  catch (Exception ex) {
					  listener.systemMessage(ex.getMessage());
				  }
				  try {
					  String allocateString = "";
					  if (plugin_config.getIntParameter("Zero New") == 1) {
						  allocateString = "Yes";
					  } else {
						  allocateString = "No";
					  }
					  String incrementalString = "";
					  if (plugin_config.getIntParameter("Enable incremental file creation") == 1) {
						  incrementalString = "Yes";
					  } else {
						  incrementalString = "No";
					  }
					  listener.systemMessage("File: Allocate space: " + allocateString + " | Incremental file creation: " + incrementalString + " | Cache size: " + plugin_config.getIntParameter("diskmanager.perf.cache.size") + "MB");
					  }
					  catch (Exception ex) {
						  listener.systemMessage(lCatchError + ex.getMessage().toString() + ")");
					  }
				  try {
					  PluginInterface dht_pi 	= AzureusCoreFactory.getSingleton().getPluginManager().getPluginInterfaceByClass( "com.aelitis.azureus.plugins.dht.DHTPlugin" );
					  DHTPlugin dhtPlugin = (DHTPlugin)dht_pi.getPlugin();
					  int dht_status = (dhtPlugin == null) ? DHTPlugin.STATUS_DISABLED	: dhtPlugin.getStatus();
					  String dHTStatus = "";
						switch (dht_status) {
						case DHTPlugin.STATUS_RUNNING:
							DHT[] dhts = dhtPlugin.getDHTs();
							if (dhts[0].getControl().getTransport().isReachable()) {
								dHTStatus = "Running (Reachable) Users: " + dhts[0].getControl().getStats().getEstimatedDHTSize();
							}
							else {
								dHTStatus = "Running (Unreachable) Users: " + dhts[0].getControl().getStats().getEstimatedDHTSize();
							}
							break;

						case DHTPlugin.STATUS_DISABLED:
							dHTStatus = "Disabled";
							break;

						case DHTPlugin.STATUS_INITALISING:
							dHTStatus = "Initialising";
							break;

						case DHTPlugin.STATUS_FAILED:
							dHTStatus = "Failed";
							break;
					}
				  int	nat_status = connection_manager.getNATStatus();
				  String nat_State = "";
				  switch (nat_status) {
				  case ConnectionManager.NAT_UNKNOWN:
					  nat_State = "Grey";
					  break;
				  case ConnectionManager.NAT_OK:
					  nat_State = "Green";
					  break;
				  case ConnectionManager.NAT_PROBABLY_OK:
					  nat_State = "Yellow";
					  break;
				  default:
					  nat_State = "Red";
				  break;
				  }
				  listener.systemMessage("Status: NAT: " + nat_State + " | DHT: " + dHTStatus);
				  }
				  catch (Exception ex) {
					  listener.systemMessage(lCatchError + ex.getMessage().toString() + ")");
				  }
				  String eLine = "";
				  try {
				  if (Constants.isWindows) {
					  Process p = Runtime.getRuntime().exec("ipconfig /all");
					  InputStream is = p.getInputStream();
					  InputStreamReader isr = new InputStreamReader(is);
					  BufferedReader br = new BufferedReader(isr);
					  String line;
					  while ((line = br.readLine()) != null) {
						  if ((line.indexOf("adapter") > -1) || (line.indexOf("Karta") > -1)) {
							  if (eLine != "") {
								  listener.systemMessage("Network: " + eLine);
								  eLine = "";
							  }			
							  if (line.indexOf("Karta") > -1)
								  eLine = eLine + "Adapter Name: " + line.substring(line.indexOf("Karta") + 8, (line.length() - 1));
							  else
								  eLine = eLine + "Adapter Name: " + line.substring(line.indexOf("adapter") + 8, (line.length() - 1));
							  }
						  else if ((line.indexOf("Description") > -1) || (line.indexOf("Beschreibung") > -1) || (line.indexOf("Opis") > -1)) {
							  eLine = eLine + " | Description: " + line.substring(line.indexOf(": ") + 2);
							  }
						  else if ((line.indexOf("Dhcp Enabled") > -1) || (line.indexOf("DHCP aktiviert") > -1) || (line.indexOf("DHCP w") > -1)){
							  eLine = eLine + " | Using DHCP: " + line.substring(line.indexOf(": ") + 2);
							  }
						  else if ((line.indexOf("IP Address") > -1) || (line.indexOf("IP-Adresse") > -1) || (line.indexOf("Adres IP") > -1)) {
							  eLine = eLine + " | IP Address: " + line.substring(line.indexOf(": ") + 2);
							  }
						  else if ((line.indexOf("Subnet Mask") > -1) || (line.indexOf("Subnetzmaske") > -1) || (line.indexOf("Maska podsieci") > -1)) {
							  eLine = eLine + " | Subnet Mask: " + line.substring(line.indexOf(": ") + 2);
							  }
						  else if ((line.indexOf("Default Gateway") > -1) || (line.indexOf("Standardgateway") > -1) || (line.indexOf("Brama domy") > -1)) {
							  eLine = eLine + " | Default Gateway: " + line.substring(line.indexOf(": ") + 2);
							  }
						  else if ((line.indexOf("DHCP Server") > -1) || (line.indexOf("DHCP-Server") > -1) || (line.indexOf("Serwer DHCP") > -1)) {			
							  eLine = eLine + " | DHCP Server: " + line.substring(line.indexOf(": ") + 2);
							  }
						  else if ((line.indexOf("DNS Servers") > -1) || (line.indexOf("DNS-Server") > -1) || (line.indexOf("Serwery DNS") > -1)) {
							  eLine = eLine + " | DNS Server: " + line.substring(line.indexOf(": ") + 2);
							  }
						  }
					  br.close();
					  if (eLine == "")
					  {
						  try {
							  eLine = "IP Address: " + InetAddress.getLocalHost ().getHostAddress() + " | Hostname: " + InetAddress.getLocalHost().getHostName();
							  }
						  catch (UnknownHostException exn) {
							  eLine = "IP Address could not be resolved";
						  }
					  }
					  listener.systemMessage("Network: " + eLine);
					  }
				  else if (Constants.isLinux) {
					  Process p = Runtime.getRuntime().exec("/sbin/ifconfig");
					  InputStream is = p.getInputStream();
					  InputStreamReader isr = new InputStreamReader(is);
					  BufferedReader br = new BufferedReader(isr);
					  String line;
					  while ((line = br.readLine()) != null) {
						  if (line.indexOf("Link encap:") > -1) {
							  if (eLine != "") {
								  listener.systemMessage("Network: " + eLine);
								  eLine = "";
							  }			
							  eLine = eLine + "Adapter Name: " + line.substring(0, line.indexOf(" "));
							  }
							
						  if (line.indexOf("Link encap:") > -1)
							  {
							  String type;
							  if (line.substring(0, line.indexOf(" ")).equals("lo"))
							  {
								  type = "Local loopback";
							  } else {
								  String tempLine = line.substring(line.indexOf("Link encap:") + 11);
								  type = tempLine.substring(0, tempLine.indexOf(" "));
							  }
							  eLine = eLine + " | Type: " + type;
							  }
						  if (line.indexOf("inet addr:") > -1) {
							  String tempLine = line.substring(line.indexOf("inet addr:") + 10);
							  eLine = eLine + " | IP Address: " + tempLine.substring(0, tempLine.indexOf(" "));
							  }
						  if (line.indexOf("Bcast:") > -1) {
							  String tempLine = line.substring(line.indexOf("Bcast:") + 6);
							  eLine = eLine + " | Broadcast: " + tempLine.substring(0, tempLine.indexOf(" "));
							  }
						  if (line.indexOf("Mask:") > -1) {
							  eLine = eLine + " | Subnet Mask: " + line.substring(line.indexOf("Mask:") + 5);
							  }
					  }
					  br.close();
					  if (eLine == "")
					  {
						  try {
							  eLine = "IP Address: " + InetAddress.getLocalHost ().getHostAddress() + " | Hostname: " + InetAddress.getLocalHost().getHostName();
							  }
						  catch (UnknownHostException exn) {
							  eLine = "IP Address could not be resolved";
						  }
					  }
					  listener.systemMessage("Network: " + eLine);
				  }
				  else if (Constants.isOSX)
				  {
					  Process p = Runtime.getRuntime().exec("/sbin/ifconfig");
					  InputStream is = p.getInputStream();
					  InputStreamReader isr = new InputStreamReader(is);
					  BufferedReader br = new BufferedReader(isr);
					  String line;
					  while ((line = br.readLine()) != null) {
						  if (line.indexOf(": flags=") > -1) {
							  if (eLine != "") {
								  listener.systemMessage("Network: " + eLine);
								  eLine = "";
							  }			
							  eLine = eLine + "Adapter Name: " + line.substring(0, line.indexOf(" ") - 1);
							  }
						  if (line.indexOf("inet ") > -1) {
							  String tempLine = line.substring(line.indexOf("inet ") + 5);
							  eLine = eLine + " | IP Address: " + tempLine.substring(0, tempLine.indexOf(" "));
							  }
						  if (line.indexOf("netmask ") > -1) {
							  String tempLine = line.substring(line.indexOf("netmask ") + 8);
							  eLine = eLine + " | Subnet Mask: " + tempLine.substring(0, tempLine.indexOf(" "));
							  }

						  if (line.indexOf("broadcast ") > -1) {
							  eLine = eLine + " | Broadcast: " + line.substring(line.indexOf("broadcast ") + 10);
							  }
					  }
					  if (eLine == "")
					  {
						  try {
							  eLine = "IP Address: " + InetAddress.getLocalHost ().getHostAddress() + " | Hostname: " + InetAddress.getLocalHost().getHostName();
							  }
						  catch (UnknownHostException exn) {
							  eLine = "IP Address could not be resolved";
						  }
					  }
					  br.close();
					  listener.systemMessage("Network: " + eLine);
				  }
				  else {
					  try {
						  listener.systemMessage("Network: IP Address: " + InetAddress.getLocalHost ().getHostAddress() + " | Hostname: " + InetAddress.getLocalHost().getHostName());
						  }
					  catch (UnknownHostException exn) {
						  listener.systemMessage("Network: IP Address: IP Address could not be resolved");
					  }
				  }
			  }
			  catch (Exception ex) {
				  try {
					  listener.systemMessage("Network: IP Address: " + InetAddress.getLocalHost ().getHostAddress() + " | Hostname: " + InetAddress.getLocalHost().getHostName());
					  }
				  catch (UnknownHostException exn) {
					  listener.systemMessage("Network: IP Address: IP Address could not be resolved");
				  }
			  }
				  try {
					  PluginInterface upnp_pi = AzureusCoreFactory.getSingleton().getPluginManager().getPluginInterfaceByClass( "com.aelitis.azureus.plugins.upnp.UPnPPlugin" );
				  String languageM = "";
				  if (Locale.getDefault().toString().equals("")) {
					  languageM = "en_US";
					  }
				  else  {
					  languageM = Locale.getDefault().toString();
					  }
				  String FLine = "";
				  String FLine2 = "";
				  if (Constants.isWindowsXP)	{
					  boolean service = true;
					  try {
						  Process p = Runtime.getRuntime().exec("REG QUERY HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess /v Start");
						  InputStream is = p.getInputStream();
						  InputStreamReader isr = new InputStreamReader(is);
						  BufferedReader br = new BufferedReader(isr);
						  String line;
						  while ((line = br.readLine()) != null) {
								  if (line.indexOf("0x4") > -1) {
									  service = false;
									  }
							  }
						  br.close();
					  } catch (Exception ex)
					  {
						  
					  }
					  if (service)
					  {
					  FLine = "XP Firewall: Enabled | ";
					  FLine2 = "XP Firewall Exceptions: Enabled | ";
					  Process p = Runtime.getRuntime().exec("REG QUERY HKLM\\System\\CurrentControlSet\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\StandardProfile");
					  InputStream is = p.getInputStream();
					  InputStreamReader isr = new InputStreamReader(is);
					  BufferedReader br = new BufferedReader(isr);
					  String line;
					  while ((line = br.readLine()) != null) {
						  if (line.indexOf("EnableFirewall") > -1) {
							  if (line.indexOf("0x0") > -1) {
								  FLine = "XP Firewall: Disabled | ";
								  }
							  else {
								  FLine = "XP Firewall: Enabled | ";
								  }
							  }
						  else if (line.indexOf("DoNotAllowExceptions") > -1)
						  {
							  if (line.indexOf("0x0") > -1) {
								  FLine2 = "XP Firewall Exceptions: Enabled | ";
								  }
							  else {
								  FLine2 = "XP Firewall Exceptions: Disabled | ";
								  }
						  }
						  }
					  br.close();
					  } else {
						  FLine = "XP Firewall: Service disabled | ";
					  }
				  }
				  if (plugin_config.getIntParameter("Plugin.UPnP.upnp.enable", 1) == 0) {
					  UPnPEnabled = "Disabled";
					  }
				  else	{
					  String info = "";
					  if ( upnp_pi != null ) {
						  info = " (" + upnp_pi.getPluginconfig().getPluginStringParameter( "plugin.info" ) + ")";
						  if (info.equals(" ()")) {
							  info = " (Unable to determine device)";
							  }
						  }
					  UPnPEnabled = "Enabled" + info;			
					  }
				  if (plugin_config.getIntParameter("User Mode", 0) == 1) {
					  UserSelectedMode = "Intermediate";
					  }
				  else if (plugin_config.getIntParameter("User Mode", 0) == 2) {
					  UserSelectedMode = "Advanced";
				  }
				  else {
					  UserSelectedMode = "Beginner";
					  }
				  listener.systemMessage("Misc: UPnP: " + UPnPEnabled + " | " + FLine + FLine2 + /*RLine + RLine2 +*/ "User Mode: " + UserSelectedMode + " | User language: " + languageM);
				  }
				  catch (Exception ex) {
					  listener.systemMessage(lCatchError + ex.getMessage().toString() + ")");
				  }
				  try {
					  String endWord = plugin_interface.getPluginDirectoryName().substring(0, plugin_interface.getPluginDirectoryName().lastIndexOf(File.separator));
					  listener.systemMessage("Plugins installed: " + plugin_manager.getPlugins().length + " | Directory: " + endWord);
				  int nb_plugins = plugin_manager.getPlugins().length;		
				  String[] plugins_names = new String[nb_plugins];
				  for(int i=0; i<nb_plugins; i++) {
					  plugins_names[i] = plugin_manager.getPlugins()[i].getPluginName();
				  }
				  Arrays.sort(plugins_names);
				  int begin = 0;
				  int nb_per_line = 15;
				  int end = nb_per_line;
				  while(begin<nb_plugins) {
					  String plugins = "";
					  for(int i=begin; i< end; i++) {			  
							  plugins += plugins_names[i] + ", ";
						  }
					  begin += nb_per_line;
					  end = Math.min(end + nb_per_line, nb_plugins);
					  if (plugins != "") {
						  listener.systemMessage("Plugins: " + plugins);
						  }
					  }
				  }
				  catch (Exception ex) {
					  listener.systemMessage(lCatchError + ex.getMessage().toString() + ")");
				  }
				  listener.systemMessage("End of Info for " + userName);
			  }
	  catch (Exception e) {
		  listener.systemMessage(lCatchError + e.getMessage().toString() + ")");
		  }
	  }

  protected void onInfo(String sourceNick, String sourceLogin, String sourceHostname, String target,int code) {
	  try {
		  String UPnPEnabled;
		  String UserSelectedMode;
		  String BindToIP = "N/A";
		  boolean bSendInfo = plugin_config.getPluginBooleanParameter(CONFIG_IRC_SEND_USER_INFO);
		  if (isChannelVoice(sourceNick) || isChannelOp(sourceNick) || sourceNick.equals(getUserName())) {
			  if(bSendInfo){
				  if (getOutgoingQueueSize() >= 3)
				  {
					  this.sendRawLine("PRIVMSG " + sourceNick + " :I'm currently busy delivering an INFO to another helper, please wait for your request to be processed. Messages in queue: " + getOutgoingQueueSize() + " | Estimated delay (seconds): " + getOutgoingQueueSize() * 3 + "");
				  }
				  try {				  
				  String patchLevel = " " + System.getProperty("sun.os.patch.level") + " ";
				  if (patchLevel == "  ") {
					  patchLevel = " ";
				  }
				  this.sendMessage(sourceNick, "Info(" + code + ") for " + getUserName() + " (" + static_id +") in " + channel +" - " + azVersion + " installed in " +  System.getProperty("user.dir") + " - " + System.getProperty("os.name") + patchLevel + System.getProperty("os.version") + " " + System.getProperty("os.arch"));
				  }
				  catch (Exception ex) {
					  this.sendMessage(sourceNick, catchError + ex.getMessage().toString() + ")");
				  }
				  if((code & REQUESTED_JAVA) != 0) {
				  try {
					  String filedir = "";
					  try {
						  try {	
							  try {
								  try {
						  File dir = new File(System.getProperty("java.home") + File.separator + ".." + File.separator + (Constants.isOSX?".."+File.separator:""));
						    String[] children = dir.list();
					  int found = 0;
					  if (children != null)
					  {
						  for (int i=0; i<children.length; i++) {
							  if (children[i].startsWith("j")) {
							  filedir = children[i] + ", " + filedir;
							  found++;
							  }
							  }
						  }
					  filedir = filedir.substring(0,filedir.length() - 3);
					  filedir = " (" + found + "): " + filedir + " |";
					  }
						  catch (Exception ex) {
							  File dir = new File(System.getProperty("java.home") + File.separator + ".." + File.separator + (Constants.isOSX?".."+File.separator:"") + (Constants.isLinux?".."+File.separator:""));
							    String[] children = dir.list();
						  int found = 0;
						  if (children != null)
						  {
							  for (int i=0; i<children.length; i++) {
								  if (children[i].startsWith("j")) {
								  filedir = children[i] + ", " + filedir;
								  found++;
								  }
								  }
							  }
						  filedir = filedir.substring(0,filedir.length() - 3);
						  filedir = " (" + found + "): " + filedir + " |";
						  }
						  }
						  catch (Exception ex) {
							  File dir = new File(System.getProperty("java.home") + File.separator + ".." + File.separator + (Constants.isOSX?".."+File.separator:""));
							    String[] children = dir.list();
						  int found = 0;
						  if (children != null)
						  {
							  for (int i=0; i<children.length; i++) {
								  filedir = children[i] + ", " + filedir;
								  found++;
								  }
							  }
						  filedir = filedir.substring(0,filedir.length() - 3);
						  filedir = " (" + found + "): " + filedir + " |";
						  }
					  } catch (Exception ex) {
						  File dir = new File(System.getProperty("java.home") + File.separator + ".." + File.separator + (Constants.isOSX?".."+File.separator:"")+ (Constants.isLinux?".."+File.separator:""));
						    String[] children = dir.list();
					  int found = 0;
					  if (children != null)
					  {
						  for (int i=0; i<children.length; i++) {
							  filedir = children[i] + ", " + filedir;
							  found++;
							  }
						  }
					  filedir = filedir.substring(0,filedir.length() - 3);
					  filedir = " (" + found + "): " + filedir + " |";
						  }
				  } catch (Exception ex) {
					  filedir = catchError + ex.getMessage().toString();
					  }
				  String JLine = "";
				  if (Constants.isWindows) {
					  try {
						  Process p = Runtime.getRuntime().exec("REG QUERY \"HKLM\\SOFTWARE\\JavaSoft\\Java Update\\Policy\\\" /s");
						  InputStream is = p.getInputStream();
						  InputStreamReader isr = new InputStreamReader(is);
						  BufferedReader br = new BufferedReader(isr);
						  String line;
						  while ((line = br.readLine()) != null) {
							  if (line.indexOf("EnableJavaUpdate") > -1) {
								  if (line.indexOf("0x0") > -1) {
									  JLine = "Updater: Disabled | ";
									  }
								  else  {
									  JLine = "Updater: Enabled | ";
									  }
								  } 
							  else if (line.indexOf("EnableAutoUpdateCheck") > -1) {
								  if (line.indexOf("0x0") > -1) {
									  JLine = JLine + "Auto update check: Disabled";
									  }
								  else	{
									  JLine = JLine + "Auto update check: Enabled";
									  }
								  }
							  }
						  br.close();
						  }
					  catch (Exception Ex) {
						  
					  }
					  }
				  this.sendMessage(sourceNick, "Java: " + System.getProperty("java.version") + " (" + System.getProperty("java.vendor") + ") installed in " + System.getProperty("java.home") + " | Installations" + filedir + " " + JLine);
				  }
				  catch (Exception ex) {
					  this.sendMessage(sourceNick, catchError + ex.getMessage().toString() + ")");
				  }
				  }
				  if((code & REQUESTED_SETTINGS) != 0) {
				  try {
				  if (! plugin_config.getStringParameter("Bind IP","").equals("")) {
					  BindToIP = plugin_config.getStringParameter("Bind IP","");
					  }
				  String portUDP = "";
				  if (plugin_config.getIntParameter("Plugin.DHT.dht.portdefault") == 1) {
					  portUDP = plugin_config.getIntParameter("TCP.Listen.Port",6881) + "";
					  } 
				  else {
					  portUDP = plugin_config.getIntParameter("Plugin.DHT.dht.port") + "";
					  }
				  String pTLine = "Disabled";
				  String pPLine = "Disabled";
				  if (plugin_config.getIntParameter("Enable.Proxy",0) == 1)
				  {
					  pTLine = "Enabled";
				  }
				  if (plugin_config.getIntParameter("Proxy.Data.Enable",0) == 1)
				  {
					  pPLine = "Enabled";
				  }
				  String enLine = "Disabled";
				  if (plugin_config.getIntParameter("network.transport.encrypted.require",0) == 1)
				  {
					  enLine = "Enabled";
					  if (plugin_config.getStringParameter("network.transport.encrypted.min_level","RC4") == "RC4")
					  {
						  enLine += " | Minimum: RC4";
					  }
					  else 
					  {
						  enLine += " | Minimum: Plain";
					  }
					  if (plugin_config.getIntParameter("network.transport.encrypted.fallback.incoming",0) == 1)
					  {
						  enLine += " | Incoming fallback: Enabled";
					  }
					  else 
					  {
						  enLine += " | Incoming fallback: Disabled";
					  }
					  if (plugin_config.getIntParameter("network.transport.encrypted.fallback.outgoing",0) == 1)
					  {
						  enLine += " | Outgoing fallback: Enabled";
					  }
					  else 
					  {
						  enLine += " | Outgoing fallback: Disabled";
					  }
				  }
				  this.sendMessage(sourceNick, "Connection: TCP Port: " + plugin_config.getIntParameter("TCP.Listen.Port",6881) + " | UDP Port: " + portUDP + " | Max connection attempts: " + plugin_config.getIntParameter("network.max.simultaneous.connect.attempts",8) + " | Bind to local IP: " + BindToIP + " | Encryption: " + enLine + " | Proxy tracker: " + pTLine + " | Proxy Data: " + pPLine);
				  }
				  catch (Exception ex) {
					  this.sendMessage(sourceNick, catchError + ex.getMessage().toString() + ")");
				  }
				  try {
				  String upSeedingOnly = "";
				  if (plugin_config.getIntParameter("enable.seedingonly.upload.rate",0) == 0) {
					  upSeedingOnly = "N/A";
					  }
				  else  {
					  upSeedingOnly = "" + plugin_config.getIntParameter("Max Upload Speed Seeding KBs", 0);
					  }
				  this.sendMessage(sourceNick, "Transfer: Max upload speed: " + plugin_config.getIntParameter("Max Upload Speed KBs", 0) + " | Max upload speed while seeding: " + upSeedingOnly + " | Max download speed: " + plugin_config.getIntParameter("Max Download Speed KBs", 0) + " | Default max upload slots: " + plugin_config.getIntParameter("Max Uploads", 4) + "");
				  }
				  catch (Exception ex) {
					  this.sendMessage(sourceNick, catchError + ex.getMessage().toString() + ")");
				  }
				  try {
				  this.sendMessage(sourceNick, "Transfer: Max connections per torrent: " + plugin_config.getIntParameter("Max.Peer.Connections.Per.Torrent", 80) + " | Max connections globally: " + plugin_config.getIntParameter("Max.Peer.Connections.Total", 400) + "");
				  }
				  catch (Exception ex) {
					  this.sendMessage(sourceNick, catchError + ex.getMessage().toString() + ")");
				  }
				  try {
					  String sameIPPeers = "Disabled";
					  String prioPiece = "Disabled";
					  String lazy = "Disabled";
					  String ignorePorts = "" + plugin_config.getStringParameter("Ignore.peer.ports","None") + "";
					  if (plugin_config.getIntParameter("Allow Same IP Peers", 0) == 1)
					  {
						  sameIPPeers = "Enabled";
					  }
					  if (plugin_config.getIntParameter("Prioritize First Piece", 0) == 1)
					  {
						  prioPiece = "Enabled";
					  }
					  if (plugin_config.getIntParameter("Use Lazy Bitfield", 0) == 1)
					  {
						  lazy = "Enabled";
					  }
					  if (ignorePorts.equals(""))
					  {
						  ignorePorts = "None";
					  }
					  this.sendMessage(sourceNick, "Transfer: Same IP peers: " + sameIPPeers + " | Prioritise first piece: " + prioPiece + " | Lazy bitfield: " + lazy + " | Ignore ports: " + ignorePorts);
					  }
					  catch (Exception ex) {
						  this.sendMessage(sourceNick, catchError + ex.getMessage().toString() + ")");
					  }
				  try {
				  String numberSeedingOnly = "";
				  if (plugin_config.getIntParameter("StartStopManager_bMaxActiveTorrentsWhenSeedingEnabled", 0) == 0) {
					  numberSeedingOnly = "N/A";
					  }
				  else  {
					  numberSeedingOnly = "" + plugin_config.getIntParameter("StartStopManager_iMaxActiveTorrentsWhenSeeding");
					  }
				  String supers = "";
				  if (plugin_config.getIntParameter("Use Super Seeding", 0) == 0)
					  supers = "Disabled";
				  else
					  supers = "Enabled";
				  this.sendMessage(sourceNick, "Queue: Max downloads: " + plugin_config.getIntParameter("max downloads", 4) + " | Max torrents: " + plugin_config.getIntParameter("max active torrents", 4) + " | Max torrents while seeding: " + numberSeedingOnly + " | Super seeding: " + supers + "");
				  }
				  catch (Exception ex) {
					  this.sendMessage(sourceNick, catchError + ex.getMessage().toString() + ")");
				  }
				  }
				  if((code & REQUESTED_ACTIVE) != 0) {
				  try {
					  Download[] managers = plugin_interface.getDownloadManager().getDownloads();
					  int forced = 0;
					  int downloading = 0;
					  int downloading_f = 0;
					  int seeding = 0;
					  int seeding_f = 0;
					  int queued = 0;
					  int queued_f = 0;
					  int erroring = 0;
					  int erroring_f = 0;
					  int preparing = 0;
					  int preparing_f = 0;
					  int stopped = 0;
					  int stopped_f = 0;
					  int ready = 0;
					  int ready_f = 0;
					  int stoping = 0;
					  int stoping_f = 0;
					  int waiting = 0;
					  int waiting_f = 0;
				      for (int i = 0; i < managers.length; i++) {
				    	  Download manager = managers[i];
				    	  if (manager.isForceStart())
				    		  forced++;
				          int state = manager.getState();
				          if (state == Download.ST_DOWNLOADING)
				          {
				            downloading++;
				            if (manager.isForceStart())
				            	downloading_f++;
				          }
				          else if (state == Download.ST_SEEDING)
				          {
				            seeding++;
				            if (manager.isForceStart())
				            	seeding_f++;
				          }
				          else if (state == Download.ST_QUEUED)
				          {
				            queued++;
				            if (manager.isForceStart())
				            	queued_f++;
				          }
				          else if (state == Download.ST_ERROR)
				          {
				        	  erroring++;
				        	  if (manager.isForceStart())
					            	erroring_f++;
					          }
				          else if (state == Download.ST_PREPARING)
				          {
				        	  preparing++;
			        	  if (manager.isForceStart())
			        		  preparing_f++;
				          }
				          else if (state == Download.ST_STOPPED)
				          {
				        	  stopped++;
			        	  if (manager.isForceStart())
				            	stopped_f++;
				          }
				          else if (state == Download.ST_READY)
				          {
				        	  ready++;
				        	  if (manager.isForceStart())
					            	ready_f++;
					          }
				          else if (state == Download.ST_STOPPING)
				          {
				        	  stoping++;
				        	  if (manager.isForceStart())
					            	stoping_f++;
					          }
				          else if (state == Download.ST_WAITING)
				          {
				        	  waiting++;
				        	  if (manager.isForceStart())
					            	waiting_f++;
					          }
				        }
				      this.sendMessage(sourceNick, "Active: Total: " + managers.length + " (F: " + forced + ") | Downloading: " + downloading + " (F: " + downloading_f + ") | Seeding: " + seeding + " (F: " + seeding_f + ") | Queued: " + queued + " (F: " + queued_f + ") | Stopped: " + stopped + " (F: " + stopped_f + ") | Ready: " + ready + " (F: " + ready_f + ") | Waiting: " + waiting + " (F: " + waiting_f + ") | Checking: " + preparing + " (F: " + preparing_f + ") | Stopping: " + stoping + " (F: " + stoping_f + ") | Errored: " + erroring + " (F: " + erroring_f + ")");
				  }
				  catch (Exception ex) {
					  this.sendMessage(sourceNick, catchError + ex.getMessage().toString() + ")");
				  }
				  }
				  if((code & REQUESTED_SETTINGS) != 0) {
				  try {
					  String allocateString = "";
					  if (plugin_config.getIntParameter("Zero New") == 1) {
						  allocateString = "Yes";
					  } else {
						  allocateString = "No";
					  }
					  String incrementalString = "";
					  if (plugin_config.getIntParameter("Enable incremental file creation") == 1) {
						  incrementalString = "Yes";
					  } else {
						  incrementalString = "No";
					  }
					  this.sendMessage(sourceNick, "File: Allocate space: " + allocateString + " | Incremental file creation: " + incrementalString + " | Cache size: " + plugin_config.getIntParameter("diskmanager.perf.cache.size") + "MB");
					  }
					  catch (Exception ex) {
						  this.sendMessage(sourceNick, catchError + ex.getMessage().toString() + ")");
					  }
				  }
				  if((code & REQUESTED_STATUS) != 0) {
				  try {
					  PluginInterface dht_pi 	= AzureusCoreFactory.getSingleton().getPluginManager().getPluginInterfaceByClass( "com.aelitis.azureus.plugins.dht.DHTPlugin" );
					  DHTPlugin dhtPlugin = (DHTPlugin)dht_pi.getPlugin();
					  String dHTStatus = "";
					  int dht_status = (dhtPlugin == null) ? DHTPlugin.STATUS_DISABLED	: dhtPlugin.getStatus();
						switch (dht_status) {
						case DHTPlugin.STATUS_RUNNING:
							DHT[] dhts = dhtPlugin.getDHTs();
							if (dhts[0].getControl().getTransport().isReachable()) {
								dHTStatus = "Running (Reachable) Users: " + dhts[0].getControl().getStats().getEstimatedDHTSize();
							}
							else {
								dHTStatus = "Running (Unreachable) Users: " + dhts[0].getControl().getStats().getEstimatedDHTSize();
							}
							break;

						case DHTPlugin.STATUS_DISABLED:
							dHTStatus = "Disabled";
							break;

						case DHTPlugin.STATUS_INITALISING:
							dHTStatus = "Initialising";
							break;

						case DHTPlugin.STATUS_FAILED:
							dHTStatus = "Failed";
							break;
					}
				  int	nat_status = connection_manager.getNATStatus();
				  String nat_State = "";
				  switch (nat_status) {
				  case ConnectionManager.NAT_UNKNOWN:
					  nat_State = "Grey";
					  break;
				  case ConnectionManager.NAT_OK:
					  nat_State = "Green";
					  break;
				  case ConnectionManager.NAT_PROBABLY_OK:
					  nat_State = "Yellow";
					  break;
				  default:
					  nat_State = "Red";
				  break;
				  }
				  this.sendMessage(sourceNick, "Status: NAT: " + nat_State + " | DHT: " + dHTStatus + "");
				  }
				  catch (Exception ex) {
					  this.sendMessage(sourceNick, catchError + ex.getMessage().toString() + ")");
				  }
				  }
				  if((code & REQUESTED_NETWORK) != 0) {
				  String eLine = "";
				  try {
				  if (Constants.isWindows) {
					  Process p = Runtime.getRuntime().exec("ipconfig /all");
					  InputStream is = p.getInputStream();
					  InputStreamReader isr = new InputStreamReader(is);
					  BufferedReader br = new BufferedReader(isr);
					  String line;
					  while ((line = br.readLine()) != null) {
						  if ((line.indexOf("adapter") > -1) || (line.indexOf("Karta") > -1)) {
							  if (eLine != "") {
								  this.sendMessage(sourceNick, "Network: " + eLine);
								  eLine = "";
							  }
							  if (line.indexOf("Karta") > -1)
								  eLine = eLine + "Adapter Name: " + line.substring(line.indexOf("Karta") + 8, (line.length() - 1)) + "";
							  else
								  eLine = eLine + "Adapter Name: " + line.substring(line.indexOf("adapter") + 8, (line.length() - 1)) + "";
							  }
						  else if ((line.indexOf("Description") > -1) || (line.indexOf("Beschreibung") > -1) || (line.indexOf("Opis") > -1)) {
							  eLine = eLine + " | Description: " + line.substring(line.indexOf(": ") + 2) + "";
							  }
						  else if ((line.indexOf("Dhcp Enabled") > -1) || (line.indexOf("DHCP aktiviert") > -1) || (line.indexOf("DHCP w") > -1)){
							  eLine = eLine + " | Using DHCP: " + line.substring(line.indexOf(": ") + 2) + "";
							  }
						  else if ((line.indexOf("IP Address") > -1) || (line.indexOf("IP-Adresse") > -1) || (line.indexOf("Adres IP") > -1)) {
							  eLine = eLine + " | IP Address: " + line.substring(line.indexOf(": ") + 2) + "";
							  }
						  else if ((line.indexOf("Subnet Mask") > -1) || (line.indexOf("Subnetzmaske") > -1) || (line.indexOf("Maska podsieci") > -1)) {
							  eLine = eLine + " | Subnet Mask: " + line.substring(line.indexOf(": ") + 2) + "";
							  }
						  else if ((line.indexOf("Default Gateway") > -1) || (line.indexOf("Standardgateway") > -1) || (line.indexOf("Brama domy") > -1)) {
							  eLine = eLine + " | Default Gateway: " + line.substring(line.indexOf(": ") + 2) + "";
							  }
						  else if ((line.indexOf("DHCP Server") > -1) || (line.indexOf("DHCP-Server") > -1) || (line.indexOf("Serwer DHCP") > -1)) {			
							  eLine = eLine + " | DHCP Server: " + line.substring(line.indexOf(": ") + 2) + "";
							  }
						  else if ((line.indexOf("DNS Servers") > -1) || (line.indexOf("DNS-Server") > -1) || (line.indexOf("Serwery DNS") > -1)) {
							  eLine = eLine + " | DNS Server: " + line.substring(line.indexOf(": ") + 2) + "";
							  }
						  }
					  br.close();
					  if (eLine == "")
					  {
						  try {
							  eLine = "IP Address: " + InetAddress.getLocalHost ().getHostAddress() + " | Hostname: " + InetAddress.getLocalHost().getHostName() + "";
							  }
						  catch (UnknownHostException exn) {
							  eLine = "IP Address could not be resolved";
						  }
					  }
					  this.sendMessage(sourceNick, "Network: " + eLine);
					  }
				  else if (Constants.isLinux) {
						  Process p = Runtime.getRuntime().exec("/sbin/ifconfig");
						  InputStream is = p.getInputStream();
						  InputStreamReader isr = new InputStreamReader(is);
						  BufferedReader br = new BufferedReader(isr);
						  String line;
						  while ((line = br.readLine()) != null) {
							  if (line.indexOf("Link encap:") > -1) {
								  if (eLine != "") {
									  this.sendMessage(sourceNick, "Network: " + eLine);
									  eLine = "";
								  }			
								  eLine = eLine + "Adapter Name: " + line.substring(0, line.indexOf(" ")) + "";
								  }
								
							  if (line.indexOf("Link encap:") > -1)
								  {
								  String type;
								  if (line.substring(0, line.indexOf(" ")).equals("lo"))
								  {
									  type = "Local loopback";
								  } else {
									  String tempLine = line.substring(line.indexOf("Link encap:") + 11);
									  type = tempLine.substring(0, tempLine.indexOf(" "));
								  }
								  eLine = eLine + " | Type: " + type + "";
								  }
							  if (line.indexOf("inet addr:") > -1) {
								  String tempLine = line.substring(line.indexOf("inet addr:") + 10);
								  eLine = eLine + " | IP Address: " + tempLine.substring(0, tempLine.indexOf(" ")) + "";
								  }
							  if (line.indexOf("Bcast:") > -1) {
								  String tempLine = line.substring(line.indexOf("Bcast:") + 6);
								  eLine = eLine + " | Broadcast: " + tempLine.substring(0, tempLine.indexOf(" ")) + "";
								  }
							  if (line.indexOf("Mask:") > -1) {
								  eLine = eLine + " | Subnet Mask: " + line.substring(line.indexOf("Mask:") + 5) + "";
								  }
						  }
						  if (eLine == "")
						  {
							  try {
								  eLine = "IP Address: " + InetAddress.getLocalHost ().getHostAddress() + " | Hostname: " + InetAddress.getLocalHost().getHostName() + "";
								  }
							  catch (UnknownHostException exn) {
								  eLine = "IP Address could not be resolved";
							  }
						  }
						  br.close();
						  this.sendMessage(sourceNick, "Network: " + eLine);
					  }
				  else if (Constants.isOSX)
				  {
					  Process p = Runtime.getRuntime().exec("/sbin/ifconfig");
					  InputStream is = p.getInputStream();
					  InputStreamReader isr = new InputStreamReader(is);
					  BufferedReader br = new BufferedReader(isr);
					  String line;
					  while ((line = br.readLine()) != null) {
						  if (line.indexOf(": flags=") > -1) {
							  if (eLine != "") {
								  this.sendMessage(sourceNick, "Network: " + eLine);
								  eLine = "";
							  }			
							  eLine = eLine + "Adapter Name: " + line.substring(0, line.indexOf(" ") - 1) + "";
							  }
						  if (line.indexOf("inet ") > -1) {
							  String tempLine = line.substring(line.indexOf("inet ") + 5);
							  eLine = eLine + " | IP Address: " + tempLine.substring(0, tempLine.indexOf(" ")) + "";
							  }
						  if (line.indexOf("netmask ") > -1) {
							  String tempLine = line.substring(line.indexOf("netmask ") + 8);
							  eLine = eLine + " | Subnet Mask: " + tempLine.substring(0, tempLine.indexOf(" ")) + "";
							  }

						  if (line.indexOf("broadcast ") > -1) {
							  eLine = eLine + " | Broadcast: " + line.substring(line.indexOf("broadcast ") + 10) + "";
							  }
					  }
					  if (eLine == "")
					  {
						  try {
							  eLine = "IP Address: " + InetAddress.getLocalHost ().getHostAddress() + " | Hostname: " + InetAddress.getLocalHost().getHostName() + "";
							  }
						  catch (UnknownHostException exn) {
							  eLine = "IP Address could not be resolved";
						  }
					  }
					  br.close();
					  this.sendMessage(sourceNick, "Network: " + eLine);
				  }
				  else 
					  {
					  try {
						  this.sendMessage(sourceNick,"Network: IP Address: " + InetAddress.getLocalHost ().getHostAddress() + " | Hostname: " + InetAddress.getLocalHost().getHostName() + "");
						  }
					  catch (UnknownHostException exn) {
						  this.sendMessage(sourceNick, "Network: IP Address: IP Address could not be resolved");
					  }
					  }
				  }
				  catch (Exception ex) {
					  try {
						  this.sendMessage(sourceNick,"Network: IP Address: " + InetAddress.getLocalHost ().getHostAddress() + " | Hostname: " + InetAddress.getLocalHost().getHostName() + "");
						  }
					  catch (UnknownHostException exn) {
						  this.sendMessage(sourceNick, "Network: IP Address: IP Address could not be resolved");
					  }
				  }
				  }
				  if((code & REQUESTED_MISC) != 0) {
				  try {
					  PluginInterface upnp_pi = AzureusCoreFactory.getSingleton().getPluginManager().getPluginInterfaceByClass( "com.aelitis.azureus.plugins.upnp.UPnPPlugin" );
				  String languageM = "";
				  if (Locale.getDefault().toString().equals("")) {
					  languageM = "en_US";
					  }
				  else  {
					  languageM = Locale.getDefault().toString();
					  }

				  String FLine = "";
				  String FLine2 = "";
				  if (Constants.isWindowsXP)	{		
					  boolean service = true;
					  try {
						  Process p = Runtime.getRuntime().exec("REG QUERY HKLM\\SYSTEM\\CurrentControlSet\\Services\\SharedAccess /v Start");
						  InputStream is = p.getInputStream();
						  InputStreamReader isr = new InputStreamReader(is);
						  BufferedReader br = new BufferedReader(isr);
						  String line;
						  while ((line = br.readLine()) != null) {
								  if (line.indexOf("0x4") > -1) {
									  service = false;
									  }
							  }
						  br.close();
					  } catch (Exception ex)
					  {
						  
					  }
					  if (service)
					  {
					  FLine = "XP Firewall: Enabled | ";
					  FLine2 = "XP Firewall Exceptions: Disabled | ";
					  Process p = Runtime.getRuntime().exec("REG QUERY HKLM\\System\\CurrentControlSet\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\StandardProfile");
					  InputStream is = p.getInputStream();
					  InputStreamReader isr = new InputStreamReader(is);
					  BufferedReader br = new BufferedReader(isr);
					  String line;
					  while ((line = br.readLine()) != null) {
						  if (line.indexOf("EnableFirewall") > -1) {
							  if (line.indexOf("0x0") > -1) {
								  FLine = "XP Firewall: Disabled | ";
								  }
							  else {
								  FLine = "XP Firewall: Enabled | ";
								  }
							  }
						  else if (line.indexOf("DoNotAllowExceptions") > -1)
						  {
							  if (line.indexOf("0x0") > -1) {
								  FLine2 = "XP Firewall Exceptions: Enabled | ";
								  }
							  else {
								  FLine2 = "XP Firewall Exceptions: Disabled | ";
								  }
						  }
						  }
					  br.close();
				  } else {
					  FLine = "XP Firewall: Service disabled | ";
				  }
			  }
				  if (plugin_config.getIntParameter("Plugin.UPnP.upnp.enable", 1) == 0) {
					  UPnPEnabled = "Disabled";
					  }
				  else	{
					  String info = "";
					  if ( upnp_pi != null ) {
						  info = " (" + upnp_pi.getPluginconfig().getPluginStringParameter( "plugin.info" ) + ")";
						  if (info.equals(" ()")) {
							  info = " (Unable to determine device)";
							  }
						  }
					  UPnPEnabled = "Enabled" + info;			
					  }
				  if (plugin_config.getIntParameter("User Mode", 0) == 1) {
					  UserSelectedMode = "Intermediate";
					  }
				  else if (plugin_config.getIntParameter("User Mode", 0) == 2) {
					  UserSelectedMode = "Advanced";
				  }
				  else {
					  UserSelectedMode = "Beginner";
					  }
				  this.sendMessage(sourceNick, "Misc: UPnP: " + UPnPEnabled + " | " + FLine + FLine2 + /*RLine + RLine2 +*/ "User Mode: " + UserSelectedMode + " | User language: " + languageM + "");
				  }
				  catch (Exception ex) {
					  this.sendMessage(sourceNick, catchError + ex.getMessage().toString() + ")");
				  }
				  }
				  if((code & REQUESTED_PLUGINS) != 0) {
				  try {
					  String endWord = plugin_interface.getPluginDirectoryName().substring(0, plugin_interface.getPluginDirectoryName().lastIndexOf(File.separator));
					  this.sendMessage(sourceNick, "Plugins installed: " + plugin_manager.getPlugins().length + " | Directory: " + endWord + "");
				  int nb_plugins = plugin_manager.getPlugins().length;		
				  String[] plugins_names = new String[nb_plugins];
				  for(int i=0; i<nb_plugins; i++) {
					  plugins_names[i] = plugin_manager.getPlugins()[i].getPluginName();
				  }
				  Arrays.sort(plugins_names);
				  int begin = 0;
				  int nb_per_line = 15;
				  int end = nb_per_line;
				  while(begin<nb_plugins) {
					  String plugins = "";
					  for(int i=begin; i< end; i++) {			  
							  plugins += plugins_names[i] + ", ";
						  }
					  begin += nb_per_line;
					  end = Math.min(end + nb_per_line, nb_plugins);
					  if (plugins != "") {
						  this.sendMessage(sourceNick, "Plugins: " + plugins + "");
						  }
					  }
				  }
				  catch (Exception ex) {
					  this.sendMessage(sourceNick, catchError + ex.getMessage().toString() + ")");
				  }
				  }
				  this.sendMessage(sourceNick, "End of Info for " + userName + "");
				  } 
			  else {
				  this.sendMessage(sourceNick, "The User does not allow sending information");
				  }
			  }
		  else {
			  this.sendMessage(sourceNick, "You do not have permission to do that (" + getChannel() + ")");
			  }
		  }
	  catch (Exception e) {
		  this.sendMessage(sourceNick, catchError + e.getMessage().toString() + ")");
		  }
	  super.onInfo(sourceNick, sourceLogin, sourceHostname, target, code);
	  }
  }